/*
   This file is part of the index package,
   Copyright (C) 1997 Harold Tay, haroldtay@pacific.net.sg

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA
 */
/* Hash table module.
 */

#include "hash.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#ifndef ARENASIZE
#define ARENASIZE 256
#endif

struct cell{
  char * key;
  char * val;
  struct cell * next;
};

/* stolen from djb */

unsigned int
string_hash_function(char * buf)
{
  unsigned int h;
  int len = strlen(buf);

  h = 5381;
  while (len) {
    --len;
    h += (h << 5);
    h ^= (unsigned int) *buf++;
  }
  return h;
}

struct table{
  unsigned int (*hash)(char *);
  int (*comp)(const char *, const char *);
  struct cell ** table;
  int tsize;
  int tused;
  struct mem{
    struct cell * free;
    struct list{
      struct cell * arena;
      int aused;
      struct list * next;
    } * list;
  } mem;
};

struct table table[HASH_MAXTABLES] = {{0,},};

static struct cell *
newcell(int ti)
{
  struct cell * p;
  register struct mem * m;

#define ZERO(p) { (p)->next = NULL; (p)->key = (p)->val = NULL; }

  m = &table[ti].mem;

  if( m->free ){
    p = m->free;
    m->free = m->free->next;
    ZERO(p);
    return(p);
  }

  if( m->list->aused < ARENASIZE ){
    p = m->list->arena + m->list->aused;
    m->list->aused++;
    ZERO(p);
    return(p);
  }
  /* m->list->aused is >= ARENASIZE */
  {
    struct list * nl = malloc(sizeof(struct list));
    p = nl->arena = malloc(ARENASIZE * sizeof(struct cell));
    nl->aused = 1;
    nl->next = m->list;
    m->list = nl;
    ZERO(p);
    return(p);
  }
}
#undef ZERO

int
hash_create(unsigned int (*hash)(char *),
            int (*compare)(const char *, const char *), int hint)
{
  int i;
  if( hint < 128 )hint = 1024;
  for(i = 0; i < HASH_MAXTABLES; i++)
    if( NULL == table[i].table ){
      table[i].hash = (hash?hash:string_hash_function);
      table[i].comp = (compare?compare:strcmp);
      table[i].tsize = hint;
      table[i].table = calloc(hint, sizeof(struct cell *));
      table[i].tused = 0;
      table[i].mem.list = calloc(1, sizeof(struct list));
      table[i].mem.list->arena =
        malloc(ARENASIZE * sizeof(struct cell));
      table[i].mem.list->aused = 0;
      table[i].mem.list->next = NULL;
      return(i);
    }
  return(-1);
}

static int
checki(int ti)
{
  if( ti < 0 || ti >= HASH_MAXTABLES || NULL == table[ti].table )
    return(-1);
  return(0);
}

int hash_size(int ti) { return(table[ti].tused); }

#define HASH(ti,key) (table[ti].hash(key)%table[ti].tsize)
#define COMP(ti,key1,key2) (table[ti].comp(key1, key2))

static void
rehash(int ti, int newsize)
{
  struct cell ** newt;
  struct cell ** t;
  struct cell ** end;
  unsigned int h;
  unsigned int (*hash)(char * key);

  t = table[ti].table;
  end = t + table[ti].tsize;
  hash = table[ti].hash;

  newt = calloc(newsize, sizeof(struct cell *));
  for(; t < end; t++){
    struct cell * hp;
    for(hp = *t; hp; ){
      struct cell * tmp;
      h = hash(hp->key)%newsize;
      tmp = newt[h];
      newt[h] = hp;
      hp = hp->next;
      newt[h]->next = tmp;
    }
  }
  free(table[ti].table);
  table[ti].table = newt;
  table[ti].tsize = newsize;
  /* Leave the other fields alone */
}

char *
hash_fetch(int ti, char * key)
{
  unsigned int h;
  struct cell * hp;

  if( -1 == checki(ti) )return((char *)-1);

  h = HASH(ti, key);
  for(hp = table[ti].table[h]; hp; hp = hp->next)
    if( 0 == COMP(ti, key, hp->key) )
      return(hp->val);
  return((char *)-1);
}

int
hash_insert(int ti, char * key, char * val)
{
  unsigned int h;
  struct cell ** t;
  struct cell * hp;

  if( -1 == checki(ti) )return(-1);

  h = HASH(ti, key);
  t = table[ti].table;
  for(hp = t[h]; hp; hp = hp->next)
    if( 0 == COMP(ti, key, hp->key) )
      return(-1);
  hp = newcell(ti);
  hp->key = key;
  hp->val = val;
  hp->next = t[h];
  t[h] = hp;
  table[ti].tused++;
  if( table[ti].tsize/table[ti].tused < 2 )
    rehash(ti, table[ti].tsize * 4);
  return(0);
}

char *
hash_replace(int ti, char * key, char * val)
{
  unsigned int h;
  struct cell ** t;
  struct cell * hp;
  char * tmp;

  if( -1 == checki(ti) )return((char *)-1);

  h = HASH(ti, key);
  t = table[ti].table;
  for(hp = t[h]; hp; hp = hp->next)
    if( 0 == COMP(ti, key, hp->key) ){
      tmp = hp->val;
      hp->val = val;
      return(tmp);
    }
  return((char *)-1);
}

int
hash_delete(int ti, char * key, char ** keyp, char ** valp)
{
  unsigned int h;
  struct cell ** t;
  struct cell * hp;
  struct cell * pre = NULL;

  if( -1 == checki(ti) )return(-1);

  h = HASH(ti, key);
  t = table[ti].table;
  for(hp = t[h]; hp; pre = hp, hp = hp->next)
    if( 0 == COMP(ti, key, hp->key) ){
      if( pre )pre->next = hp->next;
      else t[h]->next = hp->next;
      if( keyp )*keyp = hp->key;
      if( valp )*valp = hp->val;
      table[ti].tused--;

      hp->next = table[ti].mem.free;
      table[ti].mem.free = hp;

      return(0);
    }
 return(-1);
}

int
hash_destroy(int ti)
{
  struct list * list;
  struct list * next;

  if( -1 == checki(ti) )return(-1);

  /* No need to free the freelist, since they all came
     from the arena list.  Just free the arenas. */

  for(list = table[ti].mem.list; list; list = next){
    next = list->next;
    free(list->arena);
    free(list);
  }

  table[ti].mem.free = NULL;
  table[ti].mem.list = NULL;
  free(table[ti].table);
  table[ti].table = NULL;
  table[ti].tsize = table[ti].tused = 0;
  table[ti].comp = NULL;
  table[ti].hash = NULL;
  return(0);
}

int
hash_walk(int ti, int (*f)(char * key, char * val))
{
  struct cell ** t;
  struct cell ** end;
  struct cell * hp;

  if( -1 == checki(ti) )return(-1);

  end = table[ti].table + table[ti].tsize;
  for(t = table[ti].table; t < end; t++){
    int ret;
    struct cell * next;
    for(hp = *t; hp; hp = next){
      next = hp->next;
      if( (ret = (*f)(hp->key, hp->val)) != 0 )
        return(ret);
    }
  }
  return(0);
}

#undef ARENASIZE
#undef HASH
#undef COMP
